/* drivers.h
 *
 * Commonly used functions available to drivers, but not declared in
 * other header files.
 */

/*====================*/
/* Interrupt controls */
/*====================*/

/*
 * The following functions set various processor hardware priority
 * levels (spl values); each returns the previous spl value.  Splx()
 * is used to restore the previous level.  See <arm/int_hndlr.h> for
 * more details.
 */
extern int  spl0 (void);			/* minimum (user) spl */
extern int  splsoftclock (void);		/* clock software int spl */
extern int  splnet (void);			/* network software int spl */
extern int  splimp (void);			/* max. network hardware spl */
extern int  spli2c (void);			/* softint spl for i2c devs */
extern int  splbio (void);			/* block I/O hardware int spl */
extern int  spltty (void);			/* serial terminal int spl */
extern int  splhitty (void);			/* tty pseudo-DMA int spl */
extern int  splvfly (void);			/* video vertical flyback int */
extern int  splsound (void);			/* sound system IRQ  */
extern int  splclock (void);			/* hardware clock */
extern int  splhigh (void);			/* maximum spl */

extern int  splx (int old_spl);			/* revert to saved priority */

/*
 * Interrupt handler control: see also <arm/int_hndlr.h>.  This is not
 * normally required with XCB devices, except that software interrupts
 * may be declared if required.  For XCB interrupt control, see
 * <arm/xcb.h>.
 */
extern void  decl_interrupt (struct int_hndlr  *ihp, int level, unsigned type);
extern void  mask_interrupt (unsigned mask, int level, unsigned type);
extern void  allow_interrupt (unsigned mask, int level, unsigned type);
extern void  decl_softint (struct int_hndlr *ihp, int level);
extern void  raise_softint (unsigned mask);

/*====================*/
/* Process scheduling */
/*====================*/

/* Standard event synchronisation primitives */
extern int  sleep (caddr_t wait_channel, int /*sched prio+mode*/);
extern void  wakeup (caddr_t wait_channel);

/*
 * Short sleep mechanism.  Put current process to sleep at PZERO-1
 * (uninterruptible by signals) for specified number of clock ticks,
 *
 * (Clock tick rate is defined to be HZ (as a macro def); or via an
 * initialised extern int hz, which contains the same value.  Use of
 * hz is preferred since it avoids need to recompile if HZ changes. hz
 * never changes once a kernel has booted.)
 */
extern void delay (int ticks);


/*=====================*/
/* Block I/O functions */
/*=====================*/

/*
 * Biodone() is called by a block device driver on completion or
 * failure of a transfer initiated by a call to its strategy function.
 * For error conditions this can validly happen before the strategy
 * function itself returns.
 */
extern void  biodone (struct buf *bp);

/*
 * Minphys() is the function normally passed as the mincnt argument to
 * physio().  It constrains the size of an individual transfer request
 * to a predefined limit.  Although defined as returning unsigned,
 * this is a historical hangover and the function actually operates by
 * modifiying the bp->b_count field of the buffer structure if it is
 * too large: no useful result is returned.  In RISC iX 1.21, minphys
 * limits the value to 128*1024 bytes; this is to prevent user
 * programs hogging physical memory by specifying enormous transfer
 * sizes, since for each transfer all pages involved are locked in
 * memory by physio().
 */
extern unsigned minphys (struct buf *bp);

/*
 * Physio performs a raw I/O transfer.  Strat is usually the drivers's
 * normal strategy function. Bp points at a buffer structure supplied
 * by the driver (normally a static or permalloc()ed struct buf; NEVER
 * an auto variable).  Dev identifies the raw device (this is
 * anomalous but harmless in most cases: see below).  Mincnt is a
 * function which limits the number of bytes to be transferred in an
 * individual request passed to the strat function; normally minphys
 * is used for this argument.  Rw defines the direction of transfer:
 * B_READ or B_WRITE.  Uio is the structure describing the transfer
 * parameters - user address(es), byte count(s) etc (consider
 * read/readv, write/writev) - normally just passed on from the raw
 * read/write function's own uio parameter.
 *
 * The strat function is called one or more times, with bp as
 * argument.  Each transfer is of a size limited by the mincnt
 * function.  For large transfers, strat is called in linear sequence
 * on contiguous sections of each user buffer (there may be more than
 * one, consider readv/writev).  For a given read/write transfer, each
 * strategy call is discrete; physio waits for completion of one
 * request before starting the next, because there is only the one,
 * supplied, struct buf available to describe requests.  Bp->b_dev is
 * set to dev (which is actually a RAW device number) but the B_PHYS
 * flag is set in b_flags as a means of distinguishing the request
 * from normal block I/O, if required.  Drivers are at liberty to
 * supply their own alternative to minphys (see description above) as
 * the mincnt function if the transfer size limit imposed by that
 * function is unsuitable for any reason: note that to satisfy the
 * compiler the supplied function must have the same type as minphys,
 * although any returned value will be ignored.
 */
extern int physio (int (*strat)(), struct buf *bp, dev_t dev,
		   int rw, unsigned (*mincnt)(struct buf *), 
		   struct uio *uio);

/*=============================*/
/* Low level memory operations */
/*=============================*/

/*
 * Compare two arrays of bytes, each `len' long for equality: returns
 * 0 if equal, 1 if not equal.
 */
extern int  bcmp (unsigned char *s1, unsigned char *s2, int len);

/*
 * Copy an array of bytes from address `from' to address `to':
 * if overlapping, result is undefined.
 */
extern void  bcopy (caddr_t from, caddr_t to, int count);

/* Set `len' bytes from address 'addr' onwards to 0. */
extern void  bzero (caddr_t addr, int count);

/*
 * Copy string s2 to s1.  s1 must be large enough.
 * return ptr to null in s1
 */
extern char *strcpy (char *to, char *from);


/*=====================*/
/* User data transfers */
/*=====================*/

/*
 * Any of the data transfer routines declared below should only be
 * used in foreground process code, at spl 0, in a context where
 * process pre-emption is safe.  The reason for these restrictions is
 * that page faults can validly occur during kernel access to the
 * user's buffer space, but such faults (which may cause I/O to be
 * started) cannot be processed in other than foreground context.
 * Since processing of faults may involve the process being put to
 * sleep waiting for a data transfer to complete, spl protection is
 * ineffective for protecting data structures from modification under
 * interrupt: spl drops to 0.  Also, another process may be run in the
 * interim, possibly accessing the same data structures as the process
 * causing the fault.
 */

/*
 * Uiomove() moves up to `n' bytes between user buffer(s) and kernel
 * memory pointed at by cp; updates *uio for amount moved.
 * uio->uio_resid is count of bytes of original user request not yet
 * moved.  Returns 0 on success else error code (typically EFAULT).
 */
extern int uiomove (caddr_t cp, int n, enum uio_rw rw, struct uio *uio);

/*
 * Transfer a single byte (c) of a user read/readv request.  Updates
 * *uio, and returns 0 if OK, else error code (typically EFAULT).
 * uio->uio_resid MUST be greater than 0 when calling, otherwise a
 * panic will occur.
 */
extern int ureadc (int c, struct uio *uio);

/*
 * Transfer single byte of a write/writev request.  Updates *uio, and
 * returns the value (as an unsigned char: 0..255) moved from the user
 * buffer, or -1 for an error condition, in which case EFAULT should
 * be returned to the user as the error code.  uio->uio_resid MUST be
 * greater than 0 when calling, since otherwise -1 is returned but
 * this cannot be distinguished (by the caller of uwritec) from a
 * transfer error.
 */
extern int uwritec (struct uio *uio);

/*
 * Low-level user<->kernel data transfers.  Not normally required (use
 * uiomove, uwritec, ureadc instead where possible).  The only places
 * where these might be needed are in the servicing of specialised
 * ioctl commands which require more than 127 bytes of data to be
 * moved in and/or out.
 */

/*
 * Copyin copies 'len' bytes from 'src' to 'dst'.  It is assumed that
 * 'src' is in user address space.  User-space page faults may occur
 * during this call (assuming that the request was valid anyway) hence
 * it should only be used in foreground (requesting process context)
 * at spl0 (e.g. in a ioctl routine).  Returns 0 or success or an
 * error code (normally EFAULT) to be returned to the user.
 */
extern int  copyin (caddr_t src, caddr_t dst, int len);

/*
 * Like copyin except src is in kernel space, dst in user space.
 */
extern int  copyout (caddr_t src, caddr_t dst, int len);


/*===================*/
/* Timing facilities */
/*===================*/ 

/*
 * Obtain the time-of-day value from the hardware clock; this is as
 * returned to user code by gettimeofday(), and is resolved to
 * microseconds.
 */
extern void microtime (struct timeval *tvp);

/*
 * Arrange for function fun to be called in approximately nticks clock
 * ticks (see description of delay() above), being passed argument
 * arg.  The function will be called in a software interrupt context
 * at splsoftclock, and must not assume anything about the current
 * process (since there may not be one).  This is useful in
 * implementing low-frequency polling of devices which do not
 * interrupt conveniently, or for timing out a device operation
 * without holding up a process by sleeping waiting for it, e.g. using
 * delay().  The function should normally use an spl call to raise the
 * processor priority to a suitable level with respect to any hardware
 * or data structures which might be accessed via a hardware interrupt
 * occuring around the point when the timeout is triggered.  Timeout()
 * itself may safely be called from any context (including the execution
 * of fun itself, e.g. to reprime a regular polling function).
 */
void timeout (int (*fun)(caddr_t), caddr_t arg, int nticks);

/*
 * Untimeout removes the first instance of a timeout queue entry
 * (identified by the values of fun and arg) previously set up by
 * calling timeout().  It returns the number of ticks which remain
 * before the timeout would have gone off, or -1 if the entry is not
 * found.  The latter case normally implies that the timeout has
 * already been triggered.  However it is possible that the function
 * has not yet actually been called, or that it has been called and
 * returned, or that the call is currently in progress, all depending
 * on the current spl and context (consider interrupts) in relation to
 * the software interrupt context in which calling takes place.
 */
extern int untimeout (int (*fun)(caddr_t), caddr_t arg);

/*=================*/
/* Message logging */
/*=================*/

/*
 * Log a system message.
 * See <sys/syslog.h> for definitions of level parameter (LOG_EMERG..LOG_DEBUG).
 * fmt is format parameter (restricted version of library printf format), following
 * arguments as required for format.
 */
extern void log (int level, char *fmt, ...);

/*==============*/
/* System panic */
/*==============*/

/*
 * Panic() takes a single string parameter: it should tersely but
 * precisely identify the cause of a panic, in particular a driver
 * should include a suitable identifying string in any message.
 * Situations where use of panic is required should be very rare: a
 * driver should use log() and shut itself down instead, wherever
 * possible. Panicf() provides formatting facilities for more complex
 * panic messages.
 */
extern panic (char *why);
extern void panicf (char *fmt, ...);

/*===================*/
/* Memory allocation */
/*===================*/
/*
 * Permalloc() is designed to be called during driver initialisation,
 * for allocating small- to medium-sized amounts of memory (say, from
 * 16 bytes to 2-3 Kb), e.g. to local control structures.  It is NOT
 * intended for frequent allocations during normal driver operation:
 * use kmem_alloc() (declared in <sys/systm.h>) for that, if
 * necessary.  However, most common types of driver should not need to
 * do *any* memory allocation in normal operation.
 *
 * Permalloc either suceeds, returning the address of the allocated
 * memory, or else it does not return (a system panic will occur).
 */
extern caddr_t permalloc (int bytes);

/*
 * Release_chunk() should be used by a driver to free up memory known
 * not to be needed.  Most drivers should not need to call it, and
 * will use the provided memory scavenging method instead: see
 * <sys/conf.h>.
 */
extern void release_chunk(caddr_t address, int bytes);

/* EOF drivers.h */
