/*
 * $Header: /smsa/bsd:sys/include/dev/sndioctl.h:bsd.riscix1_2  1.6  $
 * $Source: /smsa/bsd:sys/include/dev/sndioctl.h: $
 *
 * Copyright (c) 1988 Acorn Computers Ltd., Cambridge, England
 *
 */

/*
 * Sound device control functions.
 */


#ifndef __dev_sndioctl_h
#define __dev_sndioctl_h

/*
 * The sound driver supports both native mode (VIDC) and Sun
 * SparcStation-compatible audio sample formats.  There are 2
 * associated devices, differing by minor device number:
 *
 *   /dev/sound
 *	is the native device, requiring samples in raw VIDC1a format:
 *	channel positioning and sample rate are fully programmable.
 *	The default parameters are 20 kHz sample rate, single mono
 *	channel, central positioning.
 *
 *   /dev/audio
 *	is the Sun SparcStation emulation interface. This outputs
 *	SparcStation-format samples at a fixed rate of 8 kHz, mono.
 *	Note that the sound playback rate is linked to the video
 *	display clock timing, and will only be precisely 8 kHz when
 *	the current screen display mode uses a 24 Mhz video clock.  On
 *	machines which support alternative clock rates (in particular,
 *	the R200 series), use of true-VGA or super-VGA screen modes
 *	will produce slightly higher output rates of 8.007 kHz and
 *	8.021 kHz, respectively.
 *
 * (SparcStation is a trademark of Sun Microsystems, Inc.)
 */
#define DEV_SOUND	0x00
#define DEV_AUDIO	0xFF

/*
 * Limit on number of channels for /dev/sound (hardware defined).  The
 * number of channels must always be a power of 2, i.e.  only 1, 2, 4
 * and 8 are valid numbers.
 */
#define	SND_MAX_CHANS	8

/*
 * Channel sample rate limits vary according to the number of channels
 * in use.  The final sample period is the channel period divided by
 * the number of channels in use, and is limited as below.  Since some
 * machines can run the video/sound clock at up to 36 MHz as opposed
 * to the original limit of 24 MHz, we limit the maximum period to
 * what you can get with the highest clock rate.  In practice this
 * should be no hardship, since 6 KHz is an awfully slow and noisy
 * sample rate!  Attempts to specific periods outside this range (via
 * ioctl (.., SNDIOCSETPARAMS, ..) will fail with EINVAL.  
 *
 * As noted earlier, sound playback rates are linked to video clock
 * timings.  The driver does the best it can to compensate (and will
 * modify timings appropriately if the screen mode is changed, even
 * while sample data is playing), but it is limited by the integer
 * clock division ratios available in the hardware.  This has the
 * consequence that with other than a 24 Mhz video clock, actual
 * sample periods may not correspond exactly to the requested rate.
 * All even period values will be OK at either 24 MHz or 36 MHz
 * (super-VGA rate), but odd period values are only exact at 24 MHz.
 * With a 25.175 MHz video clock ("true-VGA" modes) there will be some
 * inaccuracy at all period values.
 */
#define	SND_MIN_PERIOD	6
#define SND_MAX_PERIOD	((255+2)*24/36)

/*
 * Device transfer unit size.  The sound hardware is implemented using
 * a double-buffered DMA scheme.  The sound device tries to keep the
 * length of each buffer in a range such that it takes between 25 and
 * 50 milliseconds to be played.  For correct results all writes
 * should be performed in exact multiples of the buffer size bsz,
 * where bsz varies according to the final overall sample period so as
 * to maintain the buffer play time within the specified limits.  The
 * current value of bsz at any time is obtained using the SNDIOCGETSZ
 * ioctl command (see below).  Alternatively, a successful fstat()
 * call on an open file descriptor for the sound device will return
 * the value in the st_blksize field.  Below are listed the minimum
 * and maximum values for bsz.
 */
#define SND_MIN_BSZ	256
#define SND_MAX_BSZ	4096

/*
 * Note: in the current version of the driver, any write call
 * specifying a transfer length which is not a multiple of the current
 * bsz value is assumed to occur at the end of the data to be played,
 * and the supplied sound data will be padded out with `silence' data
 * up to the end of the driver's internal buffer.  If more writes do
 * actually follow immediately, there may be a gap of up to 50
 * milliseconds in the audio output before the new data is played.  (A
 * future version of the driver may handle this situation better and
 * merge the new data onto the end of the old, with no gap.)
 *
 * As a further point to note, any user buffer passed in a write() or
 * writev() call must lie on a word (4-byte) boundary; if not, the
 * call will fail with EINVAL.
 */

/* Possible positions of each channel in stereo spread */
#define SND_LEFT_100  1
#define SND_LEFT_83   2
#define SND_LEFT_67   3
#define SND_CENTRE    4
#define SND_RIGHT_67  5
#define SND_RIGHT_83  6
#define SND_RIGHT_100 7

/*
 * Description of settings of the sound device, for user control.
 * This is at a low-level mapping directly onto the hardware: the
 * driver is unconcerned about how many logically distinct sound
 * channels are being used; it is the responsibility of the
 * controlling process to determine the correct combination of period
 * and image position values according to the effective number of
 * channels.
 *
 * The period value is the hardware inter-sample period, specified in
 * microseconds (limited as specified by SND_{MIN,MAX}_PERIOD). Each
 * channel sample effectively lasts for (#channels) times longer.
 * Image position values are normally replicated in the image_pos[]
 * array to suit the number of channels in use: for 1 channel, all
 * entries should be the same; for 2 channels, all even entries must
 * be programmed with the channel 0 position, and all odd entries with
 * the channel 1 position; for 4 channels, program entries 0 and 4 for
 * channel 0, 1 and 5 for channel 2, etc; for 8 channels each entry
 * corresponds to one channel's position.
 *
 * Note that breaking these rules (e.g. writing with only two
 * effective channels of sound data but with more than 2 different
 * channel position values) is not an error (the driver cannot check,
 * since it has no knowledge of the number of channels): doing so will
 * in theory produce interesting(?) diffuse sound positioning.
 */
typedef struct snd_params
{
    unsigned int period;		/* final inter-sample time in usec */
    unsigned char image_pos[SND_MAX_CHANS];
} SNDParams, *SNDParamsRef;

/* Read current params settings */
#define SNDIOCGETPARAMS _IOR('S', 1, SNDParams)

/* Set new params for device.  Disallowed on /dev/audio* (EPERM) */
#define SNDIOCSETPARAMS _IOW('S', 2, SNDParams)

/* Calls to read and write the state of the loudspeaker control flag */
#define SNDIOCGETSPKR	_IOR('S', 3, int)
#define SNDIOCSETSPKR	_IOW('S', 4, int)

/* Call to obtain current bsz (sound buffer length) value */
#define SNDIOCGETBSZ	_IOR('S', 5, int)

#endif/*__dev_sndioctl_h*/

/* EOF sndioctl.h */
