/*
 * iecs.c
 *
 * Driver for Imaginary Expansion Card - Serial
 *
 * Copyright (C) 1989,1991 Acorn Computers Ltd.
 *
 */

#include <termios.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/syslog.h>
#include <arm/ioc.h>

#include <arm/xcb.h>
#include <arm/int_hndlr.h>

/*
 * NOTE
 *
 * THE EXAMPLE DRIVER PROVIDED IN THIS FILE DOES NOT USE THE
 * NEW KERNEL FUNCTIONS DESIGNED TO MINIMISE CHARACTER
 * OVERRUN.  IF YOU WISH TO WRITE A SERIAL LINE DRIVER THAT
 * USES THESE FUNCTIONS, PLEASE GET IN TOUCH WITH CUSTOMER
 * SUPPORT AT ACORN
 */

/* Include general specs for device driver interfaces */
#include <arm/drivers.h>

/*
 * We use (some) defs from Acorn's "es" (internal R140/R260
 * serial device) driver for the 65C51 chip, so include the
 * published header.
 */
#include <dev/acia.h>

/* and some more definitions of our own */
#include "iecs.h"

/*
 * Memory scavenging support: see <sys/conf.h> for explanation.
 */
static char iecs_data_base[] = "iecs";
static int iecs_code_base() { return 1; }

/* Maximum number of cards we will support at once */
#define MAX_CARD        4

/* macros for handling minor device numbers */
/* card number */
#define CARDNO(mindev)  ((unsigned int)(mindev) / LINES_PER_CARD)
/* line number (card-relative) */
#define LINENO(mindev)  ((unsigned int)(mindev) % LINES_PER_CARD)

/* Defaults */
#define ISPEED  B9600                   /* initial line speed */
#define IFLAGS  (EVENP|ODDP|ECHO)       /* initial line flags */

/* Derived values */
#define MAX_LINES       (MAX_CARD*LINES_PER_CARD)

#define ON      1                       /* turn acia on */
#define OFF     0                       /* turn it off */


/*
 * The following public data item points at the array of tty structures for
 * all available lines, which is allocated, by iecs_init_lo(), once the
 * number of cards (and hence lines) is known.  The contiguous layout of
 * these structures is assumed by the default select function for ttys,
 * ttselect().  Because iecs_tty is a pointer and not an actual array of
 * struct ttys, we must ourselves update the d_ttys entry in our cdevsw[]
 * entry once the memory is allocated; this is done by iecs_init_lo(). 
 */
struct tty *iecs_tty;         /* tty data structures */

struct line
{
    ACIARef        l_acia;              /* -> chip registers */
    unsigned char  l_tx_off;            /* flag for transmitter state */
    unsigned char  l_cur_dcd;           /* reflects modem status */
    unsigned char  l_cur_dsr;           /* ditto */
};

static struct card
{
    struct line    c_line[LINES_PER_CARD];
    unsigned char  c_slot;              /* physical slot for card */
    /*
     * The next field is a pointer to the card interrupt
     * control register - it is R/W, showing which acia(s)
     * is/are interrupting at any point, and controlling
     * the master card IRQ line.
     */
    volatile unsigned char  *c_irq;
    struct int_hndlr c_ih;		/* interrupt acontrol */
} *card_info[MAX_CARD];

static int n_card;                      /* #cards found */
static int n_line;                      /* #lines in total */


/* Forward refs */
int  iecs_open(), iecs_start();
static void switch_modem ();
static void  card_int ();

extern caddr_t permalloc();


iecs_init_hi (slot)
  int slot;
{
    /*
     * The XCB manager has found an iecs card in slot "slot".
     * Let's initialise it, unless we've already been told about
     * MAX_CARD cards.
     */
    int card, i, status;
    struct card *cd;
    card = n_card;
    if (card == MAX_CARD)
    {
        printf ("ignoring iecs card in slot %d\n", slot);
        return;
    }
    /* Report configuration on the console */
    printf ("iecs[%d-%d]: slot %d\n", 
            card*LINES_PER_CARD, card*(LINES_PER_CARD+1)-1,
            slot);

    /* Count up known cards */ 
    ++n_card;
    n_line += LINES_PER_CARD;           /* and available lines */

    /*
     * allocate space for info record for this card
     */
    card_info[card] = (struct card *)permalloc(sizeof(struct card));

    /* Address the record for this card */
    cd = card_info[card];

    cd->c_slot = slot;

    /* Set up card interrupt register pointer */
    cd->c_irq = CARD_IRQ(slot);
    for (i = 0; i < LINES_PER_CARD; ++i)
    {
        struct line *line = &cd->c_line[i];
        ACIARef acia = CARD_ACIA(slot, i);
        line->l_acia = acia;            /* record regs pointer */
        /* Reset the chip */
        reset_acia (acia);
        /* Read its status */
        status = acia->status;
        /*
         * Get the current state of the DCD and DSR lines of the
         * hardware interface.  Note the chip's negative logic.
         */
        line->l_cur_dcd = !(status & ACIA_STAT_DCD);
        line->l_cur_dsr = !(status & ACIA_STAT_DSR);
        /* Reset leaves the transmitter disabled - track this */
        line->l_tx_off = 1;
    }
    /* Declare our interrupt handler for this card */
    cd->c_ih.ih_fn = card_int;
    cd->c_ih.ih_farg = card;
    decl_xcb_interrupt (slot, &cd->c_ih, PRIO_TTY);

    /*
     * Finally, enable interrupts from the card.  Note that as
     * explained in iecs.h, we can just write this bit - the 
     * individual irq indicator lines are not affected.
     */
    *cd->c_irq = IR_IENABLE;
}

iecs_init_lo (slot, irqs)
  int slot, irqs;
{
    /*
     * When this routine is first called on the first round, we know
     * how many of our cards are present, and can therefore allocate
     * an appropriate amount of space for required struct ttys.
     */
    if (!irqs && iecs_tty == (struct tty *)NULL)
    {
	/*
	 * The entry for this driver in cdevsw[] should contain a 
	 * pointer to the tty structures.  However, the base address
	 * of these structures are unknown when the kernel is linked,
	 * so conf.c just uses 0 and relies on this routine to fill
	 * in the information when it is known, i.e. here and now.
	 */
	struct cdevsw *cdp, *end;

	iecs_tty = (struct tty *)permalloc(n_line * sizeof(struct tty));

	/* scan down cdevsw for this driver */
	cdp = &cdevsw[0];  end = &cdevsw[nchrdev];
	while (cdp < end && cdp->d_open != iecs_open)
	    ++cdp;

	if (cdp == end)
	    log (LOG_CRIT, "iecs: cannot find entry in cdevsw\n");
	else
	    cdp->d_ttys = iecs_tty;
    }
}

/* Card shutdown function */
iecs_shutdown (slot)
  int slot;
{
    /*
     * Since this function will be called for ALL iecs cards in
     * the machine, we can't assume that the full initialisation
     * sequence has been done for all slots, because we have ignored
     * all cards above MAX_CARD (assuming MAX_CARD < XCB_SLOTS). 
     * Hence to address a card's interrupt control register, the 
     * safest way is just to use the hardware addressing macros
     * directly, and not to bother with card_info[] etc.
     */
    int lno;
    /* Resetting the card is easy: reset each acia. */
    for (lno = 0; lno < LINES_PER_CARD; ++lno)
    {
        ACIARef acia = CARD_ACIA(slot,lno);
        reset_acia (acia);
    }
    /*
     * Then ensure it cannot interrupt.  Writing 0 to the irq control
     * register clears the master interrupt enable.
     */
    *CARD_IRQ(slot) = 0;
}


iecs_open (dev, flag)
  dev_t dev;
  int flag;
{
    struct tty *tp;
    int mindev;
    struct line *lp;
    int oldpri;

    /* Check we have the requested device installed */
    mindev = minor (dev);
    if (mindev >= n_line)
        return (ENXIO);

    /* Address the relevant tty and hardware line control structures */
    tp = &iecs_tty[mindev];
    lp = &card_info[CARDNO(mindev)]->c_line[LINENO(mindev)];

    /* Sensitive operations ahead: keep interrupts out */
    oldpri = spltty ();

    /* Set up default output start function */
    tp->t_oproc = iecs_start;
    if ((tp->t_state & TS_ISOPEN) == 0)
    {
        /*
         * First open of this device since boot or last close.
         *
         * Set up the default special characters.
         */
        ttychars (tp);
        /* Set other defaults */
        tp->t_cflag |= HUPCL;

	/* Input and output speed: always the same */
        tp->t_ispeed = ISPEED;
        tp->t_ospeed = ISPEED;
	/*
	 * The generic ttyinit routine performs further initialisation
	 * of the tty structure, based on a BSD-style flags word
	 * parameter (i.e. as passed in the sg_flags field of the
	 * TIOC[GS]ET[P] ioctl command argument). 
	 */
        ttyinit (tp, IFLAGS);
        /* Reflect software state into hardware */
        iecs_param (dev);
    }
    else if ((tp->t_cflag & _TTXCLUDE) && !issuser())
    {
        /*
         * Exclusive-access bit set, and our caller is not
         * the super-user - reject this open.
         */
        splx (oldpri);
        return (EBUSY);
    }
    /* Turn on any attached modem */
    switch_modem (dev, ON);
    /*
     * Test whether we have carrier as yet.  This may be a local line,
     * in which DCD will be wired to ON.  Our hardware is such that
     * there is no easy way to implement software-controlled local
     * operation since the DCD line to the acia chip *must* be active
     * for the chip's receiver to turn on and allow reception of data.
     * However we do honour the CLOCAL bit (in t_cflag); this isn't
     * terribly useful for any program wanting to read local data from
     * an off-line modem, but it can write data to it, and on some
     * modems it is possible (by various means) to raise DCD even when
     * no carrier is really present.
     */
    if (lp->l_cur_dcd || (tp->t_cflag & CLOCAL))
    {
        /* yep - all OK */
        tp->t_state |= TS_CARR_ON;
    }
    else
    {
        /* Go wait till carrier is present */
        tp->t_state &= ~TS_CARR_ON;
        do
        {
            tp->t_state |= TS_WOPEN;
            /* We have to sleep on the raw Q for this */
            sleep ((caddr_t) &tp->t_rawq, TTIPRI);
        } while ((tp->t_state & TS_CARR_ON) == 0);
    }
    splx (oldpri);
    /* Return via the current line-discipline's open routine */
    return (*linesw[tp->t_line].l_open) (dev, tp, flag);
}


/*
 * Last close on this tty line: shut things down.
 */
iecs_close (dev, flag)
  dev_t dev;
  int   flag;
{
    /* Address relevant tty structure */
    struct tty *tp = &iecs_tty[minor(dev)];
    int oldpri;

    /*
     * Call the line discipline close routine (at base spl).
     */
    (*linesw[tp->t_line].l_close) (tp);

    /*
     * Go to suitable priority to block interrupts while
     * adjusting hardware state.
     */
    oldpri = spltty ();
    /*
     * Ensure any line break condition is cleared: just use the
     * top level ioctl to do this.
     */
    iecs_ioctl (dev, TIOCCBRK, NULL, 0);
    /*
     * If we were stuck waiting for open, or the HUPCLS flag is set,
     * turn off ("hang up") the modem line.
     */
    if ((tp->t_cflag & HUPCL) ||
	(tp->t_state & TS_WOPEN) ||
	!(tp->t_state & TS_ISOPEN))
        switch_modem (dev, OFF);
    /*
     * Finished tweaking hardware bits now.
     */
    splx (oldpri);
    /*
     * Finally, use the standard close-down code to tidy up.
     */
    ttyclose (tp);
    return 0;
}


iecs_read (dev, uio)
  dev_t dev;
  struct uio *uio;
{
    /*
     * Process user read() request - this is trivial since the
     * line discipline code does all the work - all we have to
     * do is identify the tty from the device number, and call it.
     */
    struct tty *tp = &iecs_tty[minor(dev)];
    return ((*linesw[tp->t_line].l_read) (tp, uio));
}


iecs_write (dev, uio)
  dev_t dev;
  struct uio *uio;
{
    /* User write() request is as easy as read(). */
    struct tty *tp = &iecs_tty[minor(dev)];
    return ((*linesw[tp->t_line].l_write) (tp, uio));
}



/* Line speed control values for the hardware */
static unsigned char acia_speed[1+EXTB] =
{
    0,                                      /* unused */
    ACIA_CON_RCS | ACIA_SPEED_50,
    ACIA_CON_RCS | ACIA_SPEED_75,
    ACIA_CON_RCS | ACIA_SPEED_110,
    ACIA_CON_RCS | ACIA_SPEED_134,
    ACIA_CON_RCS | ACIA_SPEED_150,
    0,                                      /* 200 invalid */
    ACIA_CON_RCS | ACIA_SPEED_300,
    ACIA_CON_RCS | ACIA_SPEED_600,
    ACIA_CON_RCS | ACIA_SPEED_1200,
    ACIA_CON_RCS | ACIA_SPEED_1800,
    ACIA_CON_RCS | ACIA_SPEED_2400,
    ACIA_CON_RCS | ACIA_SPEED_4800,
    ACIA_CON_RCS | ACIA_SPEED_9600,
    ACIA_CON_RCS | ACIA_SPEED_19200,        /* 19200 */
    0                                       /* 38400 invalid */
};

/*
 * Ioctl call - we support a few standardised serial line
 * hardware controls, otherwise the generic code handles
 * most things.
 */
iecs_ioctl (dev, cmd, data, flag)
  dev_t dev;
  int cmd;
  caddr_t data;
  int flag;
{
    /* Identify tty structure concerned */
    int mindev = minor(dev);
    struct tty *tp = &iecs_tty[mindev];
    int status, s, ispeed, ospeed;
    ACIARef acia;

    /*
     * Try a sequence of subsidiary ioctl function calls first.  The
     * convention is for all such functions (but not this top-level one)
     * to return -1 for unrecognised ioctl commands, otherwise a status
     * code (0 or error) if they have handled it or faulted it.
     * First we try the current line discipline's ioctl entry point.
     */
    status = (*linesw[tp->t_line].l_ioctl) (tp, cmd, data, flag);
    if (status >= 0)
        return (status);

    /*
     * It wasn't recognised; try the standard ttioctl routine next.
     * Note that certain standard ioctl commands are used to change 
     * hardware parameters (particularly line speed) to arbitary values,
     * some of which we do not support.  In this case we are obliged 
     * (by POSIX) to return EINVAL, explicitly failing the ioctl call
     * because we cannot implement it precisely.  However because
     * ttioctl will modify the tty structure anyway, we must check 
     * specially for such situations *before* we call it.  Conveniently
     * there is a generic function in RISC iX 1.2 to spot this type
     * of command and tell us what the new speeds would be if ttioctl
     * were to be called.
     */
    if (ttynewspeeds (cmd, data, &ispeed, &ospeed) &&
	ospeed != B0 /* not a hang-up operation */ &&
	(ispeed != ospeed /* no mixed-speed working */ ||
	 ispeed > B38400 /* invalid speed */ || 
	 acia_speed[ispeed] == 0 /* unsupported speed */))
	return EINVAL;				/* Reject it */	

    status = ttioctl (tp, cmd, data, flag);
    if (status >= 0)
    {
	/*
	 * One of the set ttioctl handles; some of these commands
	 * involve changes to the software state which should be
	 * reflected in the device hardware: the generic function
	 * ttydevmod() (specific to RISC iX 1.2) identifies such
	 * commands.
	 */
	if (ttydevmod (cmd))
            iecs_param (dev);			/* do it now */
        return (status);
    }
    /* Address the hardware for potential changes */
    acia = card_info[CARDNO(mindev)]->c_line[LINENO(mindev)].l_acia;
    /*
     * If the command is one we process ourselves here, the hardware
     * will get modified, so raise SPL first in case it is.
     */
    s = spltty ();
    /* set default result */
    status = 0;
    switch (cmd)
    {
      case TIOCSBRK:
        /* Set BRK now (disables RTS) - worry about RTS etc. later */
        acia->command |= ACIA_TIC_NTX_RTS_BRK;
        break;

      case TIOCCBRK:
        /* Clear BRK: RTS is derived from DTR */
        cmd = acia->command & ~ACIA_CMD_TIC;
        if (cmd & ACIA_CMD_DTR)
            cmd |= ACIA_TIC_TX_RTS;
        acia->command = cmd;
        break;

      case TIOCSDTR:
        /* Sets DTR+RTS; clears any BRK */
        acia->command = ((acia->command & ~(ACIA_CMD_TIC))
                         | ACIA_TIC_TX_RTS | ACIA_CMD_DTR);
        break;

      case TIOCCDTR:
        /* Clears DTR+RTS; clears any BRK */
        acia->command &= ACIA_CMD_TIC | ACIA_CMD_DTR;
        break;

        /* To Be Done: TIOCMOD{G,S} (also TIOCM{SET,BIS,BIC,GET}?) */

      default:
        /* The command has still not been identified, we return ENOTTY. */
        status = ENOTTY;
        break;
    }
    splx (s);
    return (status);
}


iecs_param (dev)
{
    int mindev;
    struct tty  *tp;
    ACIARef acia;
    int s;
    unsigned commandreg, controlreg;

    s = spltty ();
    mindev = minor(dev);
    tp = &iecs_tty[mindev];
    acia = card_info[CARDNO(mindev)]->c_line[LINENO(mindev)].l_acia;

    /*
     * Under POSIX, setting the output speed to B0 is interpreted to
     * mean `hang up the line'. For the traditional BSD ioctl commands
     * (TIOCSETP, TIOCSETN) it is the *input* speed which is used to
     * control this, but ttioctl now takes care of the software tty
     * structure so that the driver can just follow the POSIX model.
     */
    if (tp->t_ospeed == B0)
    {   /* hang up modem */
        switch_modem (dev, OFF);
        splx (s);
        return;
    }

    /*
     * Now set up new values to be installed in the control and command
     * registers.  We know that ispeed == ospeed and that the speed is a
     * supported one, because iecs_ioctl has already checked.
     */
    controlreg = acia_speed[tp->t_ispeed];
    /*
     * Define number of stop bits: 1 or 2.
     */
    if (tp->t_cflag & CSTOPB)
        controlreg |= ACIA_CON_SBN;

    commandreg = acia->command & ~(ACIA_CMD_PMC|ACIA_CMD_PME);
    switch (tp->t_cflag & CSIZE)
    {
      case CS5: controlreg |= ACIA_WL_5; break;
      case CS6: controlreg |= ACIA_WL_6; break;
      case CS7: controlreg |= ACIA_WL_7; break;
      case CS8: controlreg |= ACIA_WL_8; break;
    }
    if (tp->t_cflag & PARENB)
    {
	commandreg |= ACIA_CMD_PME;		/* send parity bit.. */
	commandreg |= (tp->t_cflag & PARODD) ? ACIA_PMC_ODD : ACIA_PMC_EVEN;
    }
    else
    {
	commandreg |= ACIA_PMC_SPACE;		/* but PME off, so no parity bit */
    }

    /* Finally, install the settings into the hardware */
    acia->command = commandreg;
    acia->control = controlreg;
    splx (s);
}


/* Interrupt handler */

/* Forward refs: */
static void tx_int (struct tty *tp);
static void rx_int (struct tty *tp, int status, int data);
static void modem_int (struct tty *tp, struct line *lp,
		       int new_dcd, int new_dsr);

/*
 * This function is called when an interrupt occurs from on iecs card
 */
void  card_int (card)
  int card;
{
    struct card *cd = card_info[card];
    int irq = *(cd->c_irq);             /* get int req register */
    int lno;
    /* Local data and status buffering */
    unsigned char l_data  [LINES_PER_CARD];
    unsigned char l_status[LINES_PER_CARD]; 

    /*
     * To minimise delays and thereby reduce the risk of
     * character overrun, we process the interrupting lines
     * in two phases: first, read the status and any input
     * character from each interrupting acia.  Then, go do the 
     * appropriate thing for each line according to the
     * status we read.
     */
    for (lno = 0; lno < LINES_PER_CARD; ++lno)
    {
        /*
         * Check the card interrupt register, which shows
         * the interrupt status of each line chip in the 
         * corresponding bit position (0..LINES_PER_CARD-1).
         */
        if (irq & (1 << lno))
        {
            /*
             * Interrupt from Line lno is present.
             * Address the relevant acia chip.
             */
            ACIARef acia = cd->c_line[lno].l_acia;
            /* Read the chip's status and record it */
            int status = acia->status;
            l_status[lno] = status;

            /*
             * Allow for spurious interrupts (rare).
             */
            if ((status & ACIA_STAT_IRQ) == 0)
                /* ghost IRQ:  remove this bit from the irq set */
                irq &= ~(1 << lno);
            else
                /* Check for RX data ready and pick it up if so */
                if (status & ACIA_STAT_RDRF)
                    l_data[lno] = acia->data;
        }
    }
    /* Now go round again, having done the most urgent tasks */
    for (lno = 0; lno < LINES_PER_CARD; ++lno)
    {
        if (irq & (1 << lno))
        {
            struct line *line = &cd->c_line[lno];
            /*
             * Work backwards from normal to produce a minor 
             * device number from the card and line numbers.
             */
            int mindev = card * LINES_PER_CARD + lno;

            /* Hence determine the tty structure */
            struct tty *tp = &iecs_tty[mindev];

            /* Get this line's status, recorded above */
            int status = l_status[lno];

            /*
             * Compute up-to-date DCD/DSR state.
             * Note negative logic.
             */
            int chip_dcd = !(status & ACIA_STAT_DCD);
            int chip_dsr = !(status & ACIA_STAT_DSR);

            if (chip_dcd != line->l_cur_dcd || 
                chip_dsr != line->l_cur_dsr)
            {
                /* Change in DCD/DSR - go handle */
                modem_int (tp, line, chip_dcd, chip_dsr);
            }
            
            /* Check for receiver interrupt.. */
            if (status & ACIA_STAT_RDRF)
                rx_int (tp, status, l_data[lno]);

            /* Then transmitter.. */
            if (status & ACIA_STAT_TDRE)
                tx_int (tp);
        }
    }
}


/* Handle Receiver interrupt */
static void rx_int (struct tty *tp, int status, int data)
{
    int cstat;

    /*
     * Wake up anyone who might happen to have been waiting
     * for the device to open.
     */
    if ((tp->t_state & TS_ISOPEN) == 0)
    {
        wakeup ((caddr_t)&tp->t_rawq);
        if ((tp->t_state & TS_WOPEN) == 0)
            return;
    }


    /*
     * Check for errors in the received data value. 
     * Encode what we find as a character status value.
     */
    if (status & ACIA_STAT_FE)
        /*
	 * Framing error, caused by line break, speed mismatch or line
	 * noise.  The break condition is not specially distinguished by
	 * our hardware, rather by the fact that the received character
	 * value is 0; the upper levels make this decision for us, so if
	 * we had had a separate "break" indication from the chip (as
	 * some other devices give) we should explictly set the data to
	 * 0 in case the chip supplied random data in this condition.
	 */
	cstat = CS_FRMERR;
    else if (status & ACIA_STAT_PE)
	/* Parity error. */
	cstat = CS_PARERR;
    else
	/* No problem. */
	cstat = CS_OK;

    /*
     * Data loss is marked by (status & ACIA_STAT_OVRN) - however there isn't
     * much we can do about it - we might care to log it but this is likely
     * to make matters worse (especially on 9600 baud lines or faster) since
     * log() can take a significant time to execute.
     * The following code is left for perusal, but not compiled.
     */
     
#if 0
    if (status & ACIA_STAT_OVRN)
    {
        /*
         * At least one previous character (we don't know exactly how many)
         * has been lost on the wire because we weren't able to 
         * respond to an interrupt quickly enough.
         * In the following, we have to extract the minor device number
         * since we were only passed a tty pointer.
         */
        log (LOG_WARNING, "iecs%d: char overrun\n", minor(tp->t_dev));
    }
#endif

    /*
     * Pass the character and its status to upper levels via the line
     * discipline's input routine.
     */
    (*linesw[tp->t_line].l_rint) (data | cstat, tp);
}


/*
 * tx_int is called when the transmitter has completed sending 
 * one character and is ready to take the next.
 */
static void tx_int (struct tty *tp)
{
    tp->t_state &= ~TS_BUSY;            /* no longer transmitting */
    if (tp->t_line)
        (*linesw[tp->t_line].l_start) (tp);
    else
        iecs_start (tp);
}


/*
 * modem_int handles changes in the state of the DCD and DSR
 * interface lines to a modem.  Note that at the moment we
 * don't do anything with the DSR bit, however, other than keep
 * track of its state.  This routine is called at spltty.
 */
static void modem_int (struct tty *tp, struct line *lp, 
		       int new_dcd, int new_dsr)
{
    /*
     * Note: only do anything about DCD changes if we're
     * not operating in "local" mode.
     */
    if (new_dcd != lp->l_cur_dcd && 
	!tp->t_cflag & CLOCAL)
    {
        /* Carrier transition */
        if (new_dcd)
        {
            /* carrier now on */
            if ((tp->t_state & TS_CARR_ON) == 0)
                (void) (*linesw[tp->t_line].l_modem) (tp, 1);
        } 
        else
        {
            /* carrier gone away */
            if (tp->t_state & TS_CARR_ON)
                if ((*linesw[tp->t_line].l_modem) (tp, 0) == 0)
                    switch_modem (tp->t_dev, OFF);
        }
    }
    lp->l_cur_dcd = new_dcd;
    lp->l_cur_dsr = new_dsr;
}


/* 
 * iecs_start is the routine called (from the line discipline code
 * and other places) to attempt to transmit a character if this
 * is possible.  In RISC iX 1.2 it is always called at spltty, so
 * need not take special defensive action against interrupts.
 *
 *
 * The code illustrated in iecs_start uses ttystart1() to extract
 * just one character to be transmitted (see comments below).
 * Although not applicable here, there is an alternative function
 * ttystartn(), which is useful where an output device can handle
 * multiple characters in one go (e.g. if there is a hardware FIFO
 * queue for transmitted characters).
 *
 * It is defined as follows:
 *
 *    int ttystartn (struct tty *tp, char *p, int igndelays);
 *
 * It returns:
 *      -1  =>  *p is the value of the single byte to transmit
 *       0  =>  there is nothing to be transmitted
 *   n > 0  =>  tp->t_outq.c_cf is pointing at n contiguous chars
 *		to transmit (driver to remove them from the queue).
 *
 * Typical calling code looks similar to the following:
 */
#if 0 /* sample code only */
    /* Space *must* be available on the device on entry here. */
    if (space_on_device > 0)
    {
	char c;
	int nc = ttystartn (tp, &c, 0);
	if (nc < 0)
	{
	    /* send single char c to the device */
	    .....
	    /* nothing more to do (no dequeueing needed) */
	}
	else if (nc == 0)
	{
	    /* nothing to do */
	}
	else
	{
	    char *cp = tp->t_outq.c_cf;
	    int n = min (nc, space_on_device);
	    /* send n bytes at *cp onwards to the device */
	    ....
	    /* then remove these from the queue */
	    ndflush (&tp->t_outq, n);
	}
    }  
#endif
/*
 * NB Either function (ttystartn or ttystart1) should only be called
 * when there is known to be room for at least one character on the
 * output device.
 */
 
iecs_start (tp)
  struct tty *tp;
{
    /*
     * Callers do not always do complete checks on whether it is
     * actually possible to transmit, so we must do our own first.
     */
    if (!(tp->t_state & (TS_BUSY|TS_TTSTOP)))
    {
	/*
	 * There may be something to do.
	 *
	 * All that we get passed is a tty pointer, so to get at our
	 * own private information about a line, we must first extract
	 * the minor device number from tp->t_dev.
	 */
	int mindev = minor(tp->t_dev);
	/* Address the line record */
	struct line *line = &card_info[CARDNO(mindev)]->c_line[LINENO(mindev)];
	/* and the relevant acia chip. */
	ACIARef acia = line->l_acia;
	int ch;

	/*
	 * Extract a single character from the output queue, if
	 * possible.  The function ttystart1() packages up a lot of the
	 * work involved here, so it's pretty easy for us.  We pass it
	 * the tty structure pointer and a flag for whether we want to
	 * ignore character timing delay information. If this were a
	 * driver for some special interface - e.g. a software one -
	 * rather than a standard serial line, we might choose to ignore
	 * character delays, by passing 1 as second parameter.  As it
	 * is, we may well have some hardcopy terminal attached out
	 * there, so must ask ttystart1 to do what is necessary.  It
	 * usefully hides the complexities from us: as far as this
	 * driver is concerned we merely have to transmit a character if
	 * asked to do so.  If ttystart1 returns a negative result we
	 * simply stop transmitting for the moment.
	 */
	ch = ttystart1 (tp, 0);
	if (ch < 0)
	{
	    /*
	     * Nothing to send: stop interrupt on TX if not already stopped.
	     */
	    if (!line->l_tx_off)
	    {
		acia->command = ((acia->command & ~ACIA_CMD_TIC) |
				 ACIA_TIC_NTX_RTS);
		line->l_tx_off = 1;
	    }
	}
	else
	{
	    /* Re-enable transmitter if previously disabled */
	    if (line->l_tx_off) 
	    {
		acia->command = ((acia->command & ~ACIA_CMD_TIC) | 
				 ACIA_TIC_TX_RTS);
		line->l_tx_off = 0;
	    }
	    /* Just load the data into the ACIA: it will do the rest. */
	    acia->data = ch;
	    /* Record that transmitter is active */
	    tp->t_state |= TS_BUSY;
	}
    }
}


/*
 * Turn the modem on or off.  Note: we are called at spltty,
 * so no extra protection on hardware access is required.
 */
static void switch_modem (dev, flag)
  dev_t dev;
  int flag;
{
    int mindev = minor (dev);
    ACIARef acia = card_info[CARDNO(mindev)]->c_line[LINENO(mindev)].l_acia;
    if (flag == OFF)
        acia->command &= ~(ACIA_CMD_DTR|ACIA_CMD_TIC); /* stop any BRK too */
    else
        /* Set DTR+RTS */
        acia->command = ((acia->command & ~ACIA_CMD_TIC) 
                         | ACIA_TIC_TX_RTS | ACIA_CMD_DTR);
}

/*
 * Memory scavenging support.  See <sys/conf.h>
 */
static char iecs_data_end[1] = { 1 };
static int iecs_code_end() { }

struct scavenge iecs_scavenge =
{
    iecs_code_base, iecs_code_end,
    iecs_data_base, iecs_data_end,
    iecs_open, NULL,
    NULL
};

/* EOF iecs.c */
