/*
 * $Header: /smsa/bsd:sys/include/nfs/rnode.h:bsd.riscix1_2  1.7  $
 * $Source: /smsa/bsd:sys/include/nfs/rnode.h: $
 *
 * Copyright (c) 1988 Acorn Computers Ltd., Cambridge, England
 */
/*      @(#)rnode.h	1.4 88/07/15 NFSSRC4.0 from 1.21 88/02/08 SMI      */
/*	Copyright (C) 1988, Sun Microsystems Inc. */

#ifndef __nfs_rnode_h
#define __nfs_rnode_h

/*
 * Remote file information structure.
 * The rnode is the "inode" for remote files.  It contains all the
 * information necessary to handle remote file on the client side.
 */
struct rnode {
	struct rnode	*r_freef;	/* free list forward pointer */
	struct rnode	*r_freeb;	/* free list back pointer */
	struct rnode	*r_hash;	/* rnode hash chain */
	struct vnode	r_vnode;	/* vnode for remote file */
	fhandle_t	r_fh;		/* file handle */
	u_char		r_flags;	/* flags, see below */
	u_char		r_writers;	/* # NFS writes in progress to file */
	short		r_error;	/* async write error */
	daddr_t		r_lastr;	/* last block read (read-ahead) */
	short		r_owner;	/* proc index for locker of rnode */
	short		r_count;	/* number of rnode locks for r_owner */
	struct ucred	*r_cred;	/* current credentials */
	struct ucred	*r_unlcred;	/* unlinked credentials */
	char		*r_unlname;	/* unlinked file name */
	struct vnode	*r_unldvp;	/* parent dir of unlinked file */
	struct vattr	r_attr;		/* cached vnode attributes */
	struct timeval	r_attrtime;	/* time attributes become invalid */
};

#define r_size	r_attr.va_size		/* file size in bytes */

/*
 * Flags
 *
 * MT: 27-Nov-91
 * -------------
 * The RINVAL flag has been added in order to work round a nasty
 * problem with NFS reads.  If do_bio() (on behalf of nfs_strategy())
 * is processing a read operation, it used to call nfs_cache_check()
 * to verify that the file mtime returned by the server on read
 * completion was in sync with the rnode's cached mtime.  If not,
 * nfs_cache_check() called nfs_purge_caches() to force out all
 * pending writes and invalidate all buffers referring to the file.
 * However do_bio() had in its hand a pointer to the buffer, marked as
 * busy, which it had just read, and if any buffer associated with the
 * file was dirty, the action of flushing the caches would cause the
 * process to block for ever, waiting hopelessly for that buffer to
 * become non-busy.
 * Now, do_bio() instead calls nfs_cache_test() to compare the
 * returned and cached mtime attribute: it just sets the RINVAL flag
 * if they differ when they weren't expected to.  A later call on
 * nfs_purge_check(), from the rwvp() subroutine of nfs_rdwr(), then
 * forces a flush to happen if required, in circumstances where it
 * cannot cause the hang-up.
 *
 * There remains a slight loop-hole in that nfs_strategy() can be
 * called to perform a read, not only via nfs_rdwr, but also via
 * swap() when paging in an executable from an NFS filesystem.
 * Ideally, swap() should be arranged to call nfs_purge_check() in
 * this case, but it cannot since it is designed to be FS-independent
 * and just uses the strategy routine for the executable's vnode
 * without caring which that is.  To explicitly check for NFS there
 * would be distinctly dirty.  However this case is probably less
 * serious in that in-use executables are much less frequently written
 * to than other files, so the RINVAL flag is not likely to be set
 * very often.  Also, the delay in purging the client caches - up to
 * the next check on attribute validity for the rnode - is probably
 * not terribly important.  Or at least no more so than the other ways
 * in which NFS diverges from standard UNIX filesystem semantics.
 */
#define	RLOCKED		0x01		/* rnode is in use */
#define	RWANT		0x02		/* someone wants a wakeup */
#define	RATTRVALID	0x04		/* Attributes in the rnode are valid */
#define	REOF		0x08		/* EOF encountered on read */
#define	RDIRTY		0x10		/* dirty pages from write operation */
#define RINVAL		0x20		/* nfs_purge_caches() required ASAP */

/*
 * Convert between vnode and rnode
 */
#define	rtov(rp)	(&(rp)->r_vnode)
#define	vtor(vp)	((struct rnode *)((vp)->v_data))
#define	vtofh(vp)	(&(vtor(vp)->r_fh))
#define	rtofh(rp)	(&(rp)->r_fh)

/*
 * Lock and unlock rnodes.
 */
#define	RLOCK(rp) { \
	while (((rp)->r_flags & RLOCKED) && \
	    (rp)->r_owner != u.u_procp - proc) { \
		(rp)->r_flags |= RWANT; \
		(void) sleep((caddr_t)(rp), PINOD); \
	} \
	(rp)->r_owner = u.u_procp - proc; \
	(rp)->r_count++; \
	(rp)->r_flags |= RLOCKED; \
	if (((rp)->r_vnode.v_flag & VISSWAP) != 0) \
		u.u_procp->p_swlocks++; \
}

#define	RUNLOCK(rp) { \
	if (--(rp)->r_count < 0) \
		panic("RUNLOCK"); \
	if (((rp)->r_vnode.v_flag & VISSWAP) != 0) \
		u.u_procp->p_swlocks--; \
	if ((rp)->r_count == 0) { \
		(rp)->r_flags &= ~RLOCKED; \
		if ((rp)->r_flags & RWANT) { \
			(rp)->r_flags &= ~RWANT; \
			wakeup((caddr_t)(rp)); \
		} \
	} \
}
#endif/*__nfs_rnode_h*/

/* EOF rnode.h */
